(*** OPENS ********************************************************************)
open Formula;; 

(*** TYPES ********************************************************************)
type polarity = Pos | Neg | Both;;

(*** FUNCTIONS ****************************************************************)

(********************** Plaisted/Greenbaum transformation *********************)
(* associate non-atom subformula with fresh variable *)
let tseitin_var tbl f =
 if Formula.is_atom f then f else Cache.cache tbl f Formula.fresh
;;

(* cnf if [z] appears positively *)
let simp_p z = function
 | Not x     -> [[!! z; !!x]]
 | And (x,y) -> [[!! z; x]; [!!z; y]]
 | Or  (x,y) -> [[!! z; x; y]]
 | Iff (x,y) -> [[!! z; !!x; y]; [!!z; x; !!y]]
 | _ -> failwith "unknown operand occured"
;;

(* cnf if [z] appears negatively *)
let simp_n z = function
 | Not x     -> [[z; x]]
 | And (x,y) -> [[z; !!x; !! y]]
 | Or  (x,y) -> [[z; !!x]; [z; !!y]]
 | Iff (x,y) -> [[z; x; y]; [z; !!x; !!y]]
 | _ -> failwith "unknown operand occured"
;;

(* simplify according to polarity *)
let simp = function
 | Pos  -> simp_p 
 | Neg  -> simp_n
 | Both -> fun z f -> simp_p z f @ simp_n z f
;;

(* alternate polarity *)
let alt = function
 | Pos  -> Neg
 | Neg  -> Pos
 | Both -> Both
;;

(* transform unary operators *)
let rec tseitin1 ftbl vtbl acc p p1 f x z =
 let x' = tseitin_var vtbl x in
 let z' = tseitin_var vtbl z in
 tseitin ftbl vtbl (simp p z' (f x') @ acc) p1 x
 
(* transform binary operators *)
and tseitin2 ftbl vtbl acc p p1 p2 f x y z =
 let x' = tseitin_var vtbl x in
 let y' = tseitin_var vtbl y in
 let z' = tseitin_var vtbl z in
 let acc = tseitin ftbl vtbl (simp p z' (f x' y') @ acc) p1 x in
 tseitin ftbl vtbl acc p2 y

(* transform formula to cnf *)
and tseitin ftbl vtbl acc p z =
 if (Hashtbl.mem ftbl (z,p)) then acc 
 else (
  Hashtbl.add ftbl (z,p) ();
  if is_atom z then acc
  else match z with
   | Not x            -> tseitin1 ftbl vtbl acc p (alt p) (!!) x z
   | And (x,y)        -> tseitin2 ftbl vtbl acc p p p (<&>) x y z
   | Or (x,y)         -> tseitin2 ftbl vtbl acc p p p (<|>) x y z
   | Iff (x, y)       -> tseitin2 ftbl vtbl acc p Both Both (<<->>) x y z
   | _ -> failwith ("illegal case: " ^ (Formula.to_string z))
);;

let cnf = function
 | Top -> [[Fresh 0; Not(Fresh 0)]]     (*trivially satisfiable CNF*)
 | Bot -> [[Fresh 0]; [Not(Fresh 0)]]   (*trivially unsatisfiable CNF*)
 | f ->
  let ftbl = Hashtbl.create 521 in
  let vtbl = Hashtbl.create 521 in
  let p = if !Debug.tseitin then Both else Pos in
  tseitin ftbl vtbl [[tseitin_var vtbl f]] p f
;;

(*** debug output ***)
let print_cnf cnf =
 if !Debug.cnf then (
  Debug.print_time ();
  Format.printf "cnf: @[";
  List.iter (fun clause ->
   List.iter (fun lit ->
    Format.printf "%a " Formula.fprintf lit) clause;
    Format.printf "@\n"
   ) cnf;
  Format.printf "@]@\n";
 );;
